// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import bigint from 'big-integer';
import Utils from './utils';
import ed25519Point from './ed25519Point';
import { getBig, getArray } from './bnInput';
import C25519Cipher from './c25519Cipher';
import AesSherableKey from './aesSherableKey';
import SecretShare from './secretShare';
import { sign, verify, signEd, verifyEd} from './c25519Sign';
import C25519Key from './c25519Key';

export default class ed25519Key {
    /**
     * @param {import('./bnInput').BNInput} x
     * @param {ed25519Point} y
     */
    constructor(x, y) {
        this.x = getBig(x);
        this.y = y;
    }
    
    public() {
        return new ed25519Key(0, this.y);
    }

    /** @param {AesSherableKey} key */
    encryptKey(key) {
        return this.encrypt(key.toArray());
    }
    
    /** @param {Uint8Array} m */
   encrypt(m) {
        //convert here
        var priv = this.x;
        var pub = this.y.toCurve25519_Wierstrass();
        var converted_c25519_key = new C25519Key(priv, pub);
        //
        return C25519Cipher.encrypt(m, converted_c25519_key);
   }
    
    /** @param {Uint8Array|string} cipher */
    decryptKey(cipher) {
        return AesSherableKey.from(this.decrypt(cipher));
    }
    
    /** @param {Uint8Array|string} data */
    decrypt(data) {
        const cipher = fromBase64(data);
        //convert here
        var priv = this.x;
        var pub = this.y.toCurve25519_Wierstrass();
        var converted_c25519_key = new C25519Key(priv, pub);
        //
        return C25519Cipher.from(cipher).decrypt(converted_c25519_key);
    }
    
    /** @param {Uint8Array|string} data */
    partialDecrypt(data) {
        const cipher = fromBase64(data);
        //convert here
        var priv = this.x;
        var pub = this.y.toCurve25519_Wierstrass(); // keep in mind this conversion isn't needed as the partialDecrpyt function only requires key.x
        var converted_c25519_key = new C25519Key(priv, pub); 
        //
        return C25519Cipher.from(cipher).partialDecrypt(converted_c25519_key);
    }

    /**
     * @param {number} threshold 
     * @param {import('./bnInput').BNInput[]} ids
     * @param {boolean} noPublic 
     */
    share(threshold, ids, noPublic = false) {
        const nums = ids.map(id => getBig(id));
        var [,shrs] = SecretShare.shareFromIds(this.x, nums, threshold, bigint(ed25519Point.order.toString())); // possible bug. requested m but was supplied with n. i'm keeping the n
        return shrs.map(shr => ed25519Key.private(shr, noPublic));
    }

    
    /** 
     * @param {string | Uint8Array} message
     */
    sign(message) {
        var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
        return signEd(buff, this);
    }

    
    /** 
     * @param {string | Uint8Array} message
     * @param {Uint8Array} signature
     */
     
    verify(message, signature) {
        var buff = typeof message === 'string' ? Buffer.from(message, 'utf-8') : message;
        return verifyEd(buff, signature, this);
    }

    toString() {
        return Buffer.from(this.toArray()).toString("base64");
    }

    toArray() {
        const x = Buffer.from(Utils.padLeft(getArray(this.x), 32));
        return Buffer.concat([x, Buffer.from(this.y.toArray())]);
    }

    /**
     * @param {import('./bnInput').BNInput} x
     * @param {boolean} noPublic
     */
    static private(x, noPublic = false) {
        return new ed25519Key(x, noPublic ? ed25519Point.infinity : ed25519Point.g.times(x));
    }

    static generate() {
        const x = Utils.random(bigint.one, bigint((ed25519Point.order - BigInt(1)).toString()));
        return new ed25519Key(x, ed25519Point.g.times(x));
    }

    /** @param {string | Uint8Array} data */
    static from(data) {
        const buffer = typeof data === "string"
            ? Buffer.from(data, "base64") : Buffer.from(data);

        const x = getBig(buffer.slice(0, 32));
        const y = ed25519Point.from(buffer.slice(32));
        return new ed25519Key(x, y);
    }

    /** @param {string} message */
    static fromString(message) {
        return ed25519Key.from(Buffer.from(message, "base64"))
    }
}

/** 
 * @param {string|Uint8Array} data
 * @return {Uint8Array}
 */
function fromBase64(data) {
    return typeof data === "string" ? Buffer.from(data, "base64") : data;
}