// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import utils from '../utils';
import { randomPrimeRSA } from '../fastPrime';
import PaillierCipher from './paillierCipher';

export default class PaillierKey {
    get isPrivate() {
        return !(this.p.equals(BigInt.zero) || this.q.equals(BigInt.zero));
    };

    /** @param {bigInt.BigInteger} n */
    constructor(n, p = BigInt.zero, q = BigInt.zero) {
        this.n = n;
        this.p = p;
        this.q = q;
        this.g = this.n.add(1);
        this.nsquare = this.n.times(this.n);
        
        if (this.isPrivate) {
            this.psquare = this.p.times(this.p);
            this.qsquare = this.q.times(this.q);
            this.pinv = utils.primeInv(p, q);
            this.hp = utils.primeInv(PaillierCipher.evalL(this.g, p, this.psquare), p);
            this.hq = utils.primeInv(PaillierCipher.evalL(this.g, q, this.qsquare), q);
        };
    }

    static async random(size = 2048) {
        var { n, p, q } = await randomPrimeRSA(size);
        return new PaillierKey(n, p, q);
    }

    public() {
        return new PaillierKey(this.n);
    }
}
