import BN from 'bn.js';
import assert from 'assert';
import { randomBytes } from 'crypto';
import { ec } from 'elliptic';
import { getBig, getBN } from '../src/bnInput';
import SecretShare from '../src/secretShare';
import Hash from '../src/hash';

describe('tssECDSA', function () {
    it('Should sign in a threshold', function fastThresholdECDSAwithHonestMajority() {
        const ecdsa = new ec('secp256k1');

        const prvKey = ecdsa.genKeyPair();
        const pubKey = ecdsa.keyFromPublic(prvKey.getPublic().encode('hex', true), 'hex');

        const msg = Buffer.from("Hello World!", "utf8");

        const threshold = 5;
        const total = (threshold - 1) * 2 + 1;
        const ids = [...Array(total)].map((_, id) => getBN(id + 1));

        /**@type {import('elliptic').curve.base.BasePoint} */
        const g = ecdsa.g;
        const q = ecdsa.n;

        const x = prvKey.getPrivate();
        const xis = share(x, ids, threshold, q);
        const m = new BN(Hash.shaBuffer(msg), 'be').mod(q);

        //------------------------------- STEP ONE ------------------------------------------------------
        const as = ids.map(() => new BN(randomBytes(32), 'be').mod(q));
        const ks = ids.map(() => new BN(randomBytes(32), 'be').mod(q));
        const rs = ks.map(ki => g.mul(ki));

        const kijs = ks.map(ki => share(ki, ids, threshold, q));
        const aijs = as.map(ai => share(ai, ids, threshold, q));

        //------------------------------- STEP TWO ------------------------------------------------------
        const ais = ids.map((_, i) => aijs.reduce((sum, aij) => sum.add(aij[i]), new BN(0)));
        const kis = ids.map((_, i) => kijs.reduce((sum, kij) => sum.add(kij[i]), new BN(0)));

        const r = rs.reduce((sum, ri) => sum.add(ri)).getX().umod(q);
        const kais = ids.map((_, i) => kis[i].mul(ais[i]).mod(q));

        //------------------------------- STEP THREE ------------------------------------------------------
        const ka = interpolate(ids, kais, q);
        const kaInv = ka.invm(q);
        const kInvis = ais.map(ai => ai.mul(kaInv).mod(q));

        const kmis = kInvis.map(kInvi => kInvi.mul(m).mod(q));
        const kxris = kInvis.map((kInvi, i) => kInvi.mul(r).mul(xis[i]).mod(q));
        const sis = kmis.map((_, i) => kmis[i].add(kxris[i]).mod(q));

        //------------------------------- STEP FOUR ------------------------------------------------------
        const s = interpolate(ids, sis, q);
        const verification = pubKey.verify(m, {r, s});

        assert.equal(verification, true);
    });

    // Secret multiplication of k * x by k(x + r) - kr
    it('Should multiply k * x secretly', function secretReSharingForMultiplication() {
        const ecdsa = new ec('secp256k1');
        const q = ecdsa.n;

        const threshold = 3;
        const total = ((threshold - 1) * 2 + 1);
        const fullIds = [...Array(total)].map((_, id) => getBN(id + 1));
        const smallIds = [...Array(threshold)].map((_, id) => getBN(id + 1));

        const k = new BN(randomBytes(32), 'be').mod(q);
        const x = new BN(randomBytes(32), 'be').mod(q);
        const r = new BN(randomBytes(32), 'be').mod(q);
        const kx = x.mul(k).mod(q);

        const kis = share(k, fullIds, threshold, q);
        const xis = share(x, smallIds, threshold, q);
        const ris = share(r, fullIds, threshold, q);

        const xArijs = smallIds.map((_, i) => xis[i].add(ris[i]).mod(q))
            .map((xAri, i) => getLi(smallIds[i], smallIds, q).mul(xAri).mod(q))
            .map((xAri) => share(xAri, fullIds, threshold, q));

        const xAris = fullIds.map((_, i) => xArijs.reduce((sum, kxjs) => sum.add(kxjs[i]).mod(q), new BN(0)));

        const kxis = fullIds.map((_, i) => kis[i].mul(xAris[i]).add(q).sub(kis[i].mul(ris[i]).mod(q)).mod(q));

        const kxTag = interpolate(fullIds, kxis, q);

        assert.equal(kxTag.toString('hex'), kx.toString('hex'))
    });
});

/**
 * @param {BN} id
 * @param {BN[]} ids
 * @param {BN} mod
 */
function getLi(id, ids, mod) {
    const bigId = getBig(id);
    const bigIds = ids.map(id => getBig(id));
    return getBN(SecretShare.getLi(bigId, bigIds, getBig(mod)));
}

/**
 * @param {BN[]} ids
 * @param {BN[]} shares
 * @param {BN} module
 * @returns {BN}
 */
function interpolate(ids, shares, module) {
    return ids.map((id, i) => getLi(id, ids, module).mul(shares[i]).mod(module))
        .reduce((sum, ki) => sum.add(ki).mod(module));
}

/**
 * @param {BN} secret
 * @param {BN[]} ids
 * @param {number} threshold
 * @param {BN} mod
 */
function share(secret, ids, threshold, mod) {
    const x = getBig(secret);
    const n = getBig(mod);
    const nums = ids.map(id => getBig(id));
    const [, shrs] = SecretShare.shareFromIds(x, nums, threshold, n);
    return shrs.map(shr => getBN(shr));
}
